# Copyright (C) 2025 J.F.Dockes
#
# License: GPL 2.1
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""Break text in phrase based on fullstop characters"""

# Notes about breaking up text for embedding.
# - pdftohtml does not try to detect paragraphs and just generates line breaks with  <br/>
# - Many epubs have explicit <p> paragraphs, but not all. In any case, this is lost in the html to
#   text translation as <br/> and </p> yield the same text (this could be fixed).
# Other formats present similar issues.
# 
# It seems that the most reliable to detect a minimal unit of meaning is to use the sentences as
# delimited by periods.
#
# Single sentences are probably too short, so we define a target segment length and we fill this up
# as much as possible with complete sentences.
# If no periods are found at a given size limit, we break up the text at a word. If this is not
# possible, the data is deemed to not be natural text and discarded.
#
# TBD: Maybe we should overlap consecutive segments ? By 15% approximately ?


import re
import sys

def _deb(s):
    print(f"{s}", file=sys.stderr)
    
restr = r"\s+"
expr = re.compile(restr)

def dotbreak(data, segsize, prefix=""):
    sentences = []
    if isinstance(data, bytes):
        data = data.decode("utf-8", errors="replace")
    if prefix:
        prefix += ": "
        
    # Split into phrases, using the period characters.
    count=0
    segment = ""
    segments = []
    for phrase in data.split("."):
        # Replace all white space with single space chars
        phrase = expr.sub(' ', phrase).strip()
        if not phrase:
            continue
        
        #_deb(f"phrase: {phrase}")
        if len(phrase) >= segsize:
            # If phrase is very long, split it in words
            splitphr = phrase.split()
            avgwordlen = len(phrase) / len(splitphr)
            if avgwordlen > 10 or avgwordlen < 3:
                # Very long phrase and strange average word length?? Does not look like text
                continue
            # Else use the text.
            for word in splitphr:
                if len(segment) + len(word) < segsize:
                    segment += word + " "
                else:
                    segments.append(prefix + segment)
                    segment = word + " "
            segment = segment.strip() + ". "
        else:
            # Phrase needs no splitting (smaller than segment). Either add it to current or begin new
            # segment
            if len(segment) + len(phrase) < segsize:
                segment += phrase + ". "
            else:
                segments.append(prefix + segment)
                segment = phrase + ". "

    # Residual ?
    if len(segment):
        segments.append(prefix + segment)
        
    return segments


if __name__ == "__main__":
    r = dotbreak("""Edmund was fond of speaking to her of Miss Crawford, but he seemed
    to think it enough that the admiral had since been spared; and she scrupled
    to point out her own remarks to him, lest it should appear like ill-nature.
    The first actual pain which Miss Crawford occasioned her, was the consequence
    of an inclination to learn to ride, which the former caught soon after
    her being settled at Mansfield from the example of the young ladies
    at the park, and which, when Edmund's acquaintance with her increased,
    led to his encouraging the wish, and the offer of his own quiet mare
    for the purpose of her first attempts, as the best fitted for a beginner
    that either stable could furnish. No pain, no injury, however, was designed
    by him to his cousin in this offer: she was not to lose a day's exercise.
    Thisverylongwordinaphrase madeofverylongwords shoulddisappear mostcertainly dontyouthinkorwhat
    Thisverylongwordinaphrase madeofverylongwords shoulddisappear mostcertainly dontyouthinkorwhat.
    """, 100)
    for s in r:
        print(f"{s}")



#!/usr/bin/env python3
# Copyright (C) 2025 J.F.Dockes
#
# License: GPL 2.1
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import sys
import os
import getopt
from hashlib import md5

import chromadb
import ollama

from recoll import recoll
import rclconfig

from rclsem_segment import dotbreak
from slicelist import slicelist
from rclsem_common import deb, common_init, get_embedding, get_rclconfig

def update_embeddings(rcldb, collection, embedmodel, embedsegsize):

    # It is possible to configure a restricting recoll query instead of processing the whole index
    rclconf = get_rclconfig()
    rclquery = rclconf.getConfParam("sem_rclquery")
    if not rclquery:
        rclquery = "mime:*"
        
    query = rcldb.query()
    query.execute(rclquery, fetchtext=True)

    ids=[]
    for doc in query:
        # Each segment sent for embedding gets the doc file name and title as prefix to maintain
        # global context.
        prefix = os.path.splitext(doc.filename)[0] + ": " + doc.title
        # Break up the document text in segments for embedding.
        sentences = dotbreak(doc.text, embedsegsize, prefix=prefix)
        # Doc identifier
        rcludi = doc.rcludi
        deb(f"{rcludi}: Got {len(sentences)} phrases")

        if len(sentences) == 0:
            continue
        
        # We both slice and batch the chroma updates, breaking up the list of phrases by slices of
        # 100
        for sentidx, sentslice in slicelist(sentences, 100):
            # Process one slice of segments. Sentidx is the base index for the slice in the whole
            # segments list.
            deb(f"  sentidx {sentidx} len(sentslice) {len(sentslice)}")
            ids=[]
            embeddings=[]
            for i in range(len(sentslice)-1):
                # We store a segment index (slicebase+offsetinslice), which forces re-splitting at
                # query time. We could store byte offsets instead for faster query time access ?
                idx = sentidx + i
                segid = rcludi + "+" + str(idx)
                # Check if already there
                results = collection.get(ids=[segid])
                if results['ids']:
                    #deb("Already there")
                    continue
                deb(f"Adding SEGID [{segid}]\n   TEXT[{sentslice[i]}]")
                ids.append(segid)
                embeddings.append(get_embedding(sentslice[i], embedmodel))
            # Store the processed slice in chromadb
            if len(ids):
                collection.add(
                    ids=ids,
                    embeddings=embeddings,
                )


########## main

def Usage(s=""):
    print(f"{s}", file=sys.stderr)
    print(f"Usage: {os.path.basename(sys.argv[0])} [-c confdir]", file=sys.stderr)
    sys.exit(1)
    
confdir=""
try:
    options, args = getopt.getopt(sys.argv[1:], "c:")
except Exception as ex:
    Usage(str(ex))
for opt,val in options:
    if opt == "-c":
        confdir = val
    else:
        Usage(f"bad option {opt}")


rcldb, collection, embedmodel, embedsegsize = common_init(confdir)

update_embeddings(rcldb, collection, embedmodel, embedsegsize)




def slicelist(lst, slicesize):
    """slice lst into slicesize chunks and return them in an iterable way. 
    Properly take care of an incomplete last slice. We return a 2-element tuple.
    The first element is the base index for the slice (index of the first element
    in the whole list), the second element is the slice itself."""
    total = len(lst)
    base=0
    slicenum = 0
    while True:
        pos = slicenum * slicesize
        nd = min(pos+slicesize, total)
        sz = nd - pos
        if sz <= 0:
            break
        yield((pos,lst[pos:pos+sz]))
        slicenum += 1


if __name__ == "__main__":
    lst = [1,2,3,4,5,6,7,8,9,10]
    for slice in slicelist(lst, 3):
        print(f"{slice}")
    



#!/usr/bin/env python3
# Copyright (C) 2025 J.F.Dockes
#
# License: GPL 2.1
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import sys
import os
from getopt import getopt
import readline

import chromadb
import ollama

from recoll import recoll
import rclconfig

from rclsem_segment import dotbreak
from rclsem_common import deb, common_init, get_embedding


# Reranking: impossible on a cpu
RERANKING_MODEL = "dengcao/Qwen3-Reranker-4B:Q5_K_M"
def rerank_results(question, candidates):
    output = {}
    for doc in candidates:
        # Simplified logic: use the chat endpoint to ask for relevance
        prompt = f"Query: {question}\nDocument: {doc}\n" + \
            f"Is this relevant? Answer only with a score 0-1."
        deb("Chatting")
        response = ollama.chat(model=RERANKING_MODEL,
                               messages=[{'role': 'user',
                                          'content': prompt}])
        deb(f"Chatdone: {response}")
        answer = response['message']['content'].strip().lower()
        deb(f"Answer: {answer}")
        output[f"{int(float(answer)*100):03d}"] = doc

    for k in output.keys():
        print(output[k])
        break


def direct_query(rcldb, collection, embedmodel, embedsegsize, question, nres=5):
    deb(f"direct_query: question:[{question}]")
    question_embedding = get_embedding(question, embedmodel)
    #deb(f"question_embedding: {question_embedding}")
    results = collection.query(
        query_embeddings=[question_embedding],
        n_results=nres,
    )
    #deb(f"Got ids: {results['ids']}")

    # We got the ids, which are made up of a recoll rcludi and a phrase offset. Fetch the texts
    myresults = []
    for id in results["ids"][0]:
        plus = id.rfind("+")
        rcludi = id[0:plus]
        # Phrase index
        phridx = int(id[plus+1:])

        # Fetch the doc from recoll
        doc = rcldb.getDoc(rcludi)

        # We would store the texts if we were doing reranking ? Not even sure, maybe be use the
        # enlarged segment instead.
        #texts.append(doc.text)

        phrases = dotbreak(doc.text, embedsegsize)
        #deb(f"phridx {phridx} Got {len(phrases)} phrases")

        # The results are (rcludi, phraseindex, ctxbefore, segment, ctxafter)
        ctxbefore = ""
        segment = ""
        ctxafter = ""
        lines = []
        for step in range(-1, 2):
            idx = phridx + step
            if idx >= 0 and idx < len(phrases):
                if idx == phridx:
                    segment = phrases[idx]
                elif idx < phridx:
                    ctxbefore += phrases[idx]
                elif idx > phridx:
                    ctxafter += phrases[idx]
        myresults.append((rcludi, phridx, ctxbefore, segment, ctxafter))
    return myresults

   
########## main

if __name__ == "__main__":
    def printseg(begofline, seg):
        line = begofline
        lines = []
        for word in seg.split():
            line = line + " " + word
            if len(line) > 70:
                lines.append(line)
                line = begofline
        if line:
            lines.append(line)
        if lines:
            lines[-1] = lines[-1] + "."
        for line in lines:
            print(f"{line}")
            
    def print_result(rcludi, ctxbefore, segment, ctxafter):
        print(f"{rcludi}")
        printseg("   ", ctxbefore)
        printseg(" * ", segment)
        printseg("   ", ctxafter)
        
    def Usage():
        me = os.path.basename(sys.argv[0])
        print(f"Usage: {me} [-c confdir] [-n nres] [-t] [word [word ...]]", file=sys.stderr)
        print(f" -t: start a cmdtalk server", file=sys.stderr)
        print(f" -n nres: Number of results to print", file=sys.stderr)
        sys.exit(1)
        
    nres=5
    confdir=""
    try:
        options, args = getopt(sys.argv[1:], "c:n:")
    except:
        Usage()
        
    for opt,val in options:
        if opt == "-c":
            confdir = val
        elif opt == "-n":
            nres = int(val)
        else:
            Usage()
            
    # Remaining args if any are a single question to run
    question=""
    for word in args:
        question = question + " " + word
        
    rcldb, collection, embedmodel, embedsegsize = common_init(confdir)
    
    if question:
        results = direct_query(rcldb, collection, embedmodel, embedsegsize, question, nres=nres)
        for rcludi, phridx, ctxbefore, segment, ctxafter in results:
            print_result(rcludi, ctxbefore, segment, ctxafter)
    else:
        while True:
            line = input("--> ")
            if not line:
                break
            results = direct_query(rcldb, collection, embedmodel, embedsegsize, line, nres=nres)
            for rcludi, phridx, ctxbefore, segment, ctxafter in results:
                print_result(rcludi, ctxbefore, segment, ctxafter)
